/*
  Copyright (c) 2013, Matthias Schiffer <mschiffer@universe-factory.net>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include "babel.h"
#include "neigh.h"
#include "packet.h"

#include <stdlib.h>


const char *gmrf_protocol_name = "babel";
const char *gmrf_protocol_version = "experimental";


static void send_hellos(gmrf_t *gmrf, gmrf_context_t *ctx, void *arg) {
	gmrf_schedule(gmrf, send_hellos, NULL, GP_BABEL_HELLO_INTERVAL*10);

	gp_babel_send_hellos(gmrf, ctx);
}

static void send_updates(gmrf_t *gmrf, gmrf_context_t *ctx, void *arg) {
	gmrf_schedule(gmrf, send_updates, NULL, GP_BABEL_UPDATE_INTERVAL*10);

	gmrf_logf(gmrf, LOG_DEBUG, "sending periodic updates.");

	gp_babel_iface_t *iface;
	for (iface = ctx->interfaces; iface; iface = iface->next) {
		gp_babel_send_update(gmrf, ctx, iface, NULL, NULL, false);
	}
}

static void maintain_neighbours(gmrf_t *gmrf, gmrf_context_t *ctx) {
	gp_babel_iface_t *iface;
	for (iface = ctx->interfaces; iface; iface = iface->next) {
		gp_babel_neigh_t **cur, **next;
		for (cur = &iface->neighbours; *cur; cur = next) {
			gp_babel_neigh_t *neigh = *cur;
			next = &neigh->next;

			if (gp_babel_neigh_get_rxcost(gmrf, neigh) == GP_BABEL_INFINITY && gp_babel_neigh_get_txcost(gmrf, neigh) == GP_BABEL_INFINITY && !neigh->ref) {
				*cur = *next;
				next = cur;
				free(neigh);

				continue;
			}

			gmrf_logf(gmrf, LOG_DEBUG, "maintenance: %02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x[%s]: %u (rx %u/tx %u)",
				  neigh->addr.d[0], neigh->addr.d[1], neigh->addr.d[2], neigh->addr.d[3],
				  neigh->addr.d[4], neigh->addr.d[5], neigh->addr.d[6], neigh->addr.d[7],
				  gmrf_iface_get_name(gmrf, iface->gmrf_iface), gp_babel_neigh_get_cost(gmrf, neigh),
				  gp_babel_neigh_get_rxcost(gmrf, neigh), gp_babel_neigh_get_txcost(gmrf, neigh));
		}
	}
}

static void maintain_announces(gmrf_t *gmrf, gmrf_context_t *ctx) {
	gp_babel_announce_t **cur, **next;
	for (cur = &ctx->announces; *cur; cur = next) {
		gp_babel_announce_t *announce = *cur;
		next = &announce->next;

		gp_babel_announce_update(gmrf, announce);

		if (!announce->nexthops) {
			*cur = *next;
			next = cur;
			gp_babel_announce_free(gmrf, ctx, announce);

			continue;
		}

		gmrf_logf(gmrf, LOG_DEBUG, "node %04x%04x, type %04x, announce %04x (%u, seqno=%04x):",
			  ntohl(*(uint32_t*)announce->node.id), ntohl(*(uint32_t*)(announce->node.id+4)),
			  announce->type, announce->key, announce->metric.metric, announce->metric.seqno);

		gp_babel_nexthop_t *nexthop;
		for (nexthop = announce->nexthops; nexthop; nexthop = nexthop->next) {
			gp_babel_neigh_t *neigh = nexthop->neigh;

			if (!neigh) {
				gmrf_logf(gmrf, LOG_DEBUG, "        local");
				continue;
			}

			gmrf_logf(gmrf, LOG_DEBUG, "        nexthop: %02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x[%s] (%u, seqno=%04x, cost=%u%s)",
				  neigh->addr.d[0], neigh->addr.d[1], neigh->addr.d[2], neigh->addr.d[3],
				  neigh->addr.d[4], neigh->addr.d[5], neigh->addr.d[6], neigh->addr.d[7],
				  neigh->iface ? gmrf_iface_get_name(gmrf, neigh->iface->gmrf_iface) : NULL,
				  nexthop->metric_seqno.metric, nexthop->metric_seqno.seqno, gp_babel_neigh_get_cost(gmrf, neigh),
				  (nexthop == announce->selected) ? ", selected" : "");
		}
	}
}

static void maintenance(gmrf_t *gmrf, gmrf_context_t *ctx, void *arg) {
	gmrf_schedule(gmrf, maintenance, NULL, GP_BABEL_MAINTENANCE_INTERVAL*10);

	maintain_neighbours(gmrf, ctx);
	maintain_announces(gmrf, ctx);
}

gmrf_context_t* gmrf_protocol_init(gmrf_t *gmrf) {
	gmrf_logf(gmrf, LOG_INFO, "initalizing...");

	gmrf_schedule(gmrf, send_hellos, NULL, GP_BABEL_HELLO_INTERVAL*10);
	gmrf_schedule(gmrf, send_updates, NULL, GP_BABEL_UPDATE_INTERVAL*10);
	gmrf_schedule(gmrf, maintenance, NULL, 0);

	gmrf_context_t *ctx = calloc(1, sizeof(gmrf_context_t));
	gmrf_random_bytes(gmrf, &ctx->self, sizeof(gp_babel_node_id_t));

	gp_babel_announce_t *announce = gp_babel_announce_new(gmrf, ctx);
	announce->node = ctx->self;
	announce->type = 1;
	announce->key = 1337;

	announce->nexthops = announce->selected = calloc(1, sizeof(gp_babel_nexthop_t));

	return ctx;
}

void gmrf_protocol_add_iface(gmrf_t *gmrf, gmrf_context_t *ctx, gmrf_iface_t *iface) {
	gmrf_logf(gmrf, LOG_INFO, "interface `%s' added.", gmrf_iface_get_name(gmrf, iface));

	gp_babel_iface_t *b_iface = calloc(1, sizeof(gp_babel_iface_t));
	b_iface->gmrf_iface = iface;

	b_iface->next = ctx->interfaces;
	ctx->interfaces = b_iface;
}

void gmrf_protocol_handle_packet(gmrf_t *gmrf, gmrf_context_t *ctx, gmrf_iface_t *iface, const gmrf_addr_t *source, const void *data, size_t len) {
	const gp_babel_packet_t *packet = data;

	if (len < sizeof(gp_babel_packet_t) || len < gp_babel_packet_size(packet)) {
		gmrf_logf(gmrf, LOG_DEBUG, "received short packet.");
		return;
	}

	gp_babel_iface_t *b_iface = gp_babel_get_iface(ctx, iface);
	if (!b_iface)
		return;

	gp_babel_handle_packet(gmrf, ctx, b_iface, source, packet);
}
